"""
Name: ACSNI-data
Author: Chinedu A. Anene, Phd
"""

import pandas as pd
from sklearn.preprocessing import MinMaxScaler
import numpy as np
import random, string, sys


def get_col_names(x):
    """
    Get names of numeric columns

    Parameters
    ----------
    x: Matrix
    """
    return x.select_dtypes([np.number]).columns

def get_row_index(x):
    """
    Get name of the row index.
    Throw error if there are more than categorical column.

    Parameters
    ----------
    x: Matrix
    """
    id_name = x.select_dtypes([np.object]).columns

    if len(id_name) != 1:
        sys.exit("ERROR: Make sure there just one ID column")

    return id_name

def parse_exp(x):
    """
    Process the expression file

    Parameters
    ----------
    x: matrix of expression

    Returns
    -------
    res: results matrix
    """

    x1 = get_row_index(x)
    x2 = get_col_names(x)
    y = pd.DataFrame()

    if len(x1) != 1:
        sys.exit("ERROR: Please, check the file and try again")

    y['name'] = x[x1[0]]
    y = pd.concat([y, x[x2]], axis=1)
    return y

def remove_unexpressed(df):
    """
    Remove unexpressed genes

    Parameters
    ----------
    df: Full expression matrix

    Returns
    -------
    res: Filtered expression matrix
    """

    x = df.select_dtypes(include=["number", "float", "int"]).copy()
    nc = x.shape[1]
    x['count'] = x.T[x.T == 0].count()
    res = df[x['count'] <= nc]
    res.reset_index(drop=True, inplace=True)
    return res

def filter_uninformative(df, mad_f):
    """
    Filter uninformative genes

    Parameters
    ----------
    df: Expression matrix
    mad_f: Minimum median absolute deviation

    Returns
    -------
    temp: Filtered expression matrix
    """

    i_gene_len = len(df)
    if i_gene_len == 0:
        sys.exit("ERROR: Include column with gene name in the matrix")
    else:
        x = df.select_dtypes(include=["number"]).copy()
        row_deviance = x.sub(x.median(axis=1), axis=0)
        mad_row = np.abs(row_deviance).median(axis=1)
        temp = df[mad_row >= mad_f]

        if temp.shape[0] <= (i_gene_len/10):
            sys.exit(">90% uninformative genes. Please, "
                     "check the expression matrix or reduce -m. "
                     "Use pre-filtered matrix if appropriate")
        else:
            temp.reset_index(drop=True, inplace=True)
    return temp

def get_scaled_values(file_ss):
    """
    Scale the expression on samples

    Parameters
    ----------
    file_ss: Filtered expression matrix

    Returns
    -------
    res: Scaled expression matrix
    """

    cols = get_col_names(file_ss)
    id_name = get_row_index(file_ss)
    scale = MinMaxScaler(feature_range=(0, 1))
    x = pd.DataFrame(scale.fit_transform(file_ss[cols]))
    out = pd.DataFrame(np.array(file_ss[id_name]))
    out.columns = ["name"]
    res = pd.concat([out, x], axis=1)
    return res

def gene_sets(prior):
    """
    Process the prior matrix

    Parameters
    ----------
    prior: Gene set matrix of 1s and 0s

    Returns
    -------
    gene_set: Dictionary of prior matrix
    """

    cols = get_col_names(prior)
    id_name = get_row_index(prior)
    prior = prior.rename(columns={id_name[0]: "name"})

    gene_set = {}
    for i in cols:
        gene_set[i] = pd.DataFrame(prior[prior[i] == 1]["name"])
    return gene_set

def scale_first(input_file, d_list):
    """
    Subset gene set expression

    Parameters
    ----------
    input_file: Filtered, scaled expression matrix
    d_list: Single gene set names

    Returns
    -------
    x: Gene set expression
    y: Gene names
    """

    temp = pd.merge(get_scaled_values(input_file), d_list, how="inner", on="name")
    print("Geneset with {} genes in the expression".format(temp.shape[0]))

    if temp.shape[0] <= 11:
        sys.exit("Fewer than 12 genes in the set, use the f 2 option "
                 "to run existing method or check expression matrix")
    else:
        cols = get_col_names(temp)
        x = temp[cols]
        x = x.reindex(columns=sorted(x.columns))
        x = np.array(x)
        y = temp["name"]
    return x, y

def scale_second(input_file):
    """
    Process expression for prediction

    Parameters
    ----------
    input_file: Complete expression matrix

    Returns
    -------
    y: Gene names
    x: Expression values
    """

    cols = get_col_names(input_file)
    id_name = get_row_index(input_file)
    x = input_file[cols]
    x = x.reindex(columns=sorted(x.columns))
    y = input_file[id_name]
    return y, x

def get_model_weights(w, y_train):
    """
    Prepare optional weight matrix

    Parameters
    ----------
    w: weight matrix
    y_train: Training gene names (Gene set)

    Returns
    -------
    w_nn1: Formatted weight matrix
    """

    nn = pd.DataFrame(y_train)
    w_nn = pd.merge(nn, w, how="left", on="name")
    w_nn["w"] = w_nn["w"].fillna(0)
    w_nn1 = w_nn["w"]
    return w_nn1

def name_generator(l=6, s=string.ascii_uppercase + string.digits):
    """
    Function to generate random names for saving files.
    :param l: Length of the string.
    :param s: Where to sample from.
    :return:
    """
    return ''.join(random.choice(s) for _ in range(l))


"""
Name: ACSNI-network
Author: Chinedu A. Anene, Phd
"""

from ACSNI.dat import get_scaled_values
import pandas as pd
from sklearn.linear_model import LinearRegression
from scipy.stats import beta
import numpy as np
from ACSNI.dat import get_col_names


class NetScore:
    """
    Class for network construction and scoring.
    """

    def __init__(self, x, w, p):
        """
        Network Class.

        Parameters:
            x: Input matrix
            w: Weights to merge and calculate
            p: probability threshold
        """
        self.x = x
        self.w = w
        self.p = p
        self.__un = None
        self.__ud = None

    def __str__(self):
        return "Network construction class"

    def __beta_params(self, x):
        """
        Estimate the parameters of beta distribution.

        Parameters
        ----------
        x: Index to get unit network from data.

        Returns
        -------
        estimates: the beta and alpha
        """

        mu, var = self.__un[x].mean(), self.__un[x].var()
        al = ((1 - mu) / var - 1 / mu) * mu ** 2
        be = al * (1 / mu - 1)
        return al, be

    def __unit_network(self):
        """
        Construct unit wise regulatory networks.

        """

        exp1 = get_scaled_values(self.x)
        w1 = self.w.select_dtypes(include=["number", "float", "int"]).copy()
        exp_set = exp1.select_dtypes(include=["number", "float", "int"]).copy().T

        self.__un = exp1[["name"]]
        self.__ud = exp1[["name"]]
        for i in range(w1.shape[1]):
            x = np.array(w1.iloc[:, i]).reshape((-1, 1))

            lm_g = []
            lm_d = []
            for t in range(exp_set.shape[1]):
                y = np.array(exp_set.iloc[:, t]).reshape((-1, 1))
                lm_r = LinearRegression(n_jobs=-1, fit_intercept=True, copy_X=False)
                lm_r.fit(x, y)
                lm_g.append(lm_r.score(x, y))
                lm_d.append(float(lm_r.coef_))

            lm_g = pd.DataFrame(lm_g, columns=[w1.columns.values[i]])
            lm_d = pd.DataFrame(lm_d, columns=[w1.columns.values[i]])

            self.__un = pd.concat([self.__un, lm_g], axis=1)
            self.__ud = pd.concat([self.__ud, lm_d], axis=1)
        return

    def __score_class(self):
        """
        Call interactions {"P", "B} and
        counts them for global prediction.

        """

        nn = get_col_names(self.__un)
        for i in nn:
            a, b = self.__beta_params(i)
            self.__un[i] = 1 - beta.cdf(self.__un[i], a, b)
            self.__un[i] = np.where(self.__un[i] <= self.p, "P", "B")

        ae_col = [col for col in self.__un if col.startswith("AE")]
        filter_ae = self.__un[ae_col]
        self.__un["Predicted"] = filter_ae.iloc[:].eq("P").sum(axis=1).to_frame()
        return

    def fit(self):
        self.__unit_network()
        self.__score_class()
        return

    def save_net(self, s, i, nn):
        """
        Save the unit network with the directions.

        Parameters
        ----------
        i: Gene set name
        s: Boot number
        nn: Input name

        """
        self.__un.to_csv("N{}_{}_{}".format(s, i, nn), index=False)
        self.__ud.to_csv("D{}_{}_{}".format(s, i, nn), index=False)
        return

    def get_predicted(self, t):

        if t == 1:
            op = self.__un
        elif t == 0:
            op = self.__un[[col for col in self.__un.columns if col in ["name", "Predicted"]]]

        else:
            op = 0
        return op





"""
Name: ACSNI-model
Author: Chinedu A. Anene, Phd
"""
from sklearn.decomposition import PCA, NMF
import warnings
from sklearn import metrics
from tensorflow.keras.models import Model, load_model
from tensorflow.keras.layers import Input, Dense
from tensorflow.keras.callbacks import ModelCheckpoint
from tensorflow.keras import regularizers, optimizers
import numpy as np
from pandas import DataFrame as Dff
import pandas as pd
from ACSNI.dat import name_generator



class DimRed:
  """
  Class for quadruple dimension reduction.
  """

  def __init__(self, x, w, p):
      """
      Dimension reduction class.

      Parameters:
          x: Input matrix (np array)
          w: Weights to adjustment for ae
          p: latent dimension for ae
      """
      self.x = x
      self.w = w
      self.p = p
      self.pca = None
      self.nmf = None
      self.ae = None
      self.__reduced = None
      self.__pcanmf = None
      self.__median = None
      self.__ael1 = None
      self.__ael2 = None
      self.__a = None
      self.__r = 15
      self.__a1 = 0.04
      self.__a2 = 0.80
      self.__scorer = metrics.explained_variance_score
      self.__run_id = name_generator(6)

  def __str__(self):
      return "Quadruple dimension reduction class"

  def __repr__(self):
      return "\n" + self.__str__()

  def __get_score(self, model, y):
      """
      Determine level of explained variance

      """
      prediction = model.inverse_transform(model.transform(y))
      return self.__scorer(y, prediction)

  def l_med(self):
      self.__median = pd.DataFrame(np.median(self.x.T, axis=1))
      self.__median = self.__median.add_prefix('MEDIAN_' + self.__run_id + '_')
      return


  def lde(self):
      """
      Decompose with PCA and NMF

      """
      self.pca = PCA(n_components=0.95)
      self.pca.fit(self.x)
      pc_weights = pd.DataFrame(self.pca.components_.T)
      pc_weights = pc_weights.add_prefix('PCA_' + self.__run_id + '_')

      opti_rank = []
      #
      warnings.filterwarnings('ignore')
      for k in range(2, self.__r):
          nmf = NMF(n_components=k, max_iter=1000).fit(self.x)
          score_it = self.__get_score(nmf, self.x)
          opti_rank.append(score_it)
          if score_it >= 0.95:
              break


      self.nmf = NMF(n_components=len(opti_rank) + 1, max_iter=10000)
      self.nmf.fit(self.x)
      warnings.resetwarnings()
      #

      nmf_weights = pd.DataFrame(self.nmf.components_.T)
      nmf_weights = nmf_weights.add_prefix('NMF_' + self.__run_id + '_')

      self.__pcanmf = pd.concat([pc_weights, nmf_weights], axis=1)

      return

  def __de4ae(self, y):
      """
      Estimate optimal dimension for AE,
        based on Bahadur and Paffenroth 2020, IEEE

      """
      s_x = y.copy()

      for t in range(s_x.shape[0]):
          s_x.iloc[t, :] = np.sort(np.array(s_x.iloc[t, :]))[::-1]

      svp = np.sort(s_x.mean())[::-1]
      svp_sum = svp.sum()
      alg1 = sum(svp / svp_sum > self.__a1)
      alg2 = 0

      temp = (svp_sum * self.__a2) / 1
      temp2 = 0

      for i in range(len(svp)):
          temp2 += svp[i]
          alg2 += 1

          if temp2 >= temp:
              break
      return int((alg1 + alg2) / 2)

  def __aer(self, nc):
      """
      Build model structure

      """
      input_layer = Input(shape=(nc,), name="input")
      encoder = Dense(self.__a, activation="relu",
                      kernel_initializer="glorot_uniform",
                      activity_regularizer=regularizers.l1_l2(1e-16, 1e-9),
                      name="enco1")(input_layer)

      encoder = Dense(self.__a // 2, activation="relu", name="code")(encoder)

      decoder = Dense(self.__a, activation="sigmoid", name="deco1")(encoder)
      decoder = Dense(nc, activation="sigmoid", name="output")(decoder)

      self.ae = Model(inputs=input_layer, outputs=decoder)
      self.ae.compile(optimizer=optimizers.RMSprop(learning_rate=1e-3),
                           loss='mean_squared_error', metrics=['mse'])
      return

  def mud_ae(self):
      """
      Model training and output.

      """
      nam = self.__run_id + "_model.h5"

      if self.p == 0:
          self.__a = self.x.shape[0] * 50 // 100
          check1 = ModelCheckpoint(filepath="est_" + nam, verbose=0,
                                   save_best_only=True)

          self.__aer(nc=self.x.shape[1])
          self.ae.fit(self.x, self.x, sample_weight=self.w,
                       epochs=500, shuffle=True, batch_size=15,
                       validation_data=(self.x, self.x), callbacks=[check1],
                       verbose=0)

          estimate = load_model("est_" + nam)
          code_est = Dff(estimate.get_layer("enco1").get_weights()[0])
          self.__a = self.__de4ae(code_est)
          print("The optimal number of dimension is {}".format(self.__a))

      else:
          self.__a = self.p

      check = ModelCheckpoint(filepath=nam, verbose=0, save_best_only=True)

      self.__aer(nc=self.x.shape[1])
      self.ae.fit(self.x, self.x, sample_weight=self.w,
                epochs=3000, shuffle=True, batch_size=15,
                validation_data=(self.x, self.x), callbacks=[check],
                verbose=0)

      final = load_model(nam)
      self.__ael1 = Dff(final.get_layer("enco1").get_weights()[0])
      self.__ael1 = self.__ael1.add_prefix('AE_' + self.__run_id + '_')

      self.__ael2 = Dff(final.get_layer("code").get_weights()[0])
      self.__ael2["run"] = nam

      self.__ael2.to_csv("code_{}.csv".format(self.__run_id), index=False)
      return

  def fit(self):
      """
      Fit quadruple dimension reduction {Median, PCA, NMF, AE[DE]}

      """
      self.l_med()
      self.lde()
      self.mud_ae()
      self.__reduced = pd.concat([self.__ael1,
                                  self.__pcanmf,
                                  self.__median],
                                 axis=1)
      return

  def get_reduced(self):
      """
      Get reduced dimension

      """
      return self.__reduced

  def get_aede(self):
      return self.__a

  def add_reduced_row(self, y):
      self.__reduced["ID"] = y
      self.__pcanmf["ID"] = y
      self.__ael1["ID"] = y
      return




"""
ACSNI-derive
Author: Faraz Khan, PhD

"""
import sys
import os
import pandas as pd
import numpy as np
from scipy.stats import ttest_ind
from ACSNI import dat, dbs
from ACSNI.corep import main_model_prediction


def parse_biotype(x):
    """
    Process the biotype file.

    Parameters
    ----------
    x: matrix of bio type

    Returns
    -------
    y: processed file
    """

    if x.shape[1] != 2:
        sys.exit("ERROR: Biotype file should have two columns of "
                 "gene names and biotype")

    x1 = len(set(x.iloc[:, 0]))
    x2 = len(set(x.iloc[:, 1]))
    y = pd.DataFrame()

    if x1 == x2:
        sys.exit("ERROR: Please, check the file and try again")

    elif x1 > x2:
        y['name'] = x.iloc[:, 0]
        y['biotype'] = x.iloc[:, 1]

    else:
        y['name'] = x.iloc[:, 1]
        y['biotype'] = x.iloc[:, 0]

    return y

def parse_cor(x):
    """
    Process the correlation file.

    Parameters
    ----------
    x: matrix of correlation

    Returns
    -------
    res: results matrix
    """

    if x.shape[1] != 2:
        sys.exit("ERROR: Correlation file should have two columns "
                 "of gene names and correlation R.")

    x1 = dat.get_row_index(x)
    x2 = dat.get_col_names(x)

    y = pd.DataFrame()

    if len(x1) != 1:
        sys.exit("ERROR: Please, check the file and try again")


    y['name'] = x[x1[0]]
    y['cor'] = x[x2[0]]

    return y

def cont_method(x, y):
    """
    Control methods {supervised}.

    Parameters
    ----------
    x: matrix
    y: target name

    Returns
    -------
    res: results matrix
    """

    xx = x.copy().T
    cols = dat.get_col_names(xx)
    xx["group"] = pd.cut(xx[y], bins=[np.min(xx[y]),
                                                   np.median(xx[y]),
                                                   np.max(xx[y])],
                            labels=["Low", "High"], include_lowest=True)

    res =  pd.DataFrame()
    for t in cols:
        x1 = np.array(xx[xx["group"] == "High"][t])
        y1 = np.array(xx[xx["group"] == "Low"][t])
        _, de = ttest_ind(x1, y1)

        de = np.array([de])
        frame1 = pd.DataFrame(de, index=[t])
        res = pd.concat([res, frame1], axis=0)
    res.rename(columns={0: "de.p"}, inplace=True)
    return res


def merge_minimal_derive(x):
    """
    Merge the minimal output results

    Parameters
    ----------
    x: Dictionary of results

    Returns
    -------
    end: Completion
    """

    mse = x[0]
    for i in x[1:]:
        mse = pd.merge(mse, i, how="outer", on="name")
    return mse

def get_ascni(prior_m, expression_m, mad, p, s, a, nn=".csv"):
    """
    Function to call ACSNI from derive.

    Parameters
    ----------
    prior_m: Prior matrix
    expression_m: Expression matrix
    mad: Mad cut
    p: percent
    s: seed
    a: p_value threshold for prediction call
    nn: The suffix to save output files

    Returns
    -------
    results

    """
    print("Searching")

    run_info = {
        "m": mad, "b": 1, "c": a,
        "p": p, "f": 0, "i": nn, "t": None,
        "w": None, "s": s}

    re = list()
    for i in prior_m:
        in_p = prior_m[i].shape[0] * p // 100
        res = main_model_prediction(inp=expression_m,
                                        d_list=prior_m[i],
                                        gi=i, lp=in_p, s=s,
                                        run=run_info)

        re.append(res)

    re = merge_minimal_derive(re)
    return re

def get_cor(exp_m, cbtype, goi, madf, cort, corf, biotypefilter=False,
            biotypef=False,docor=False):
    """
    Function to correlate gene of interest against the genes of the same biotype

    Parameters
    ----------
    exp_m: Expression matrix
    biotypef: Gene Biotype file
    cbtype: Biotype of interest
    goi: Gene of interest
    madf: MAD threshold
    cort: Correlation threshold
    biotypefilter: If biotype file and filter parameters provided
    docor: Use modified correlation file provided
    corf: Modified correlation file to be used for the genration of prior 1.

    Returns
    -------
    gcor: Correlation matrix

    """

    if biotypefilter:
        bsub = biotypef[biotypef['biotype'].isin([cbtype])]
        esub = exp_m[exp_m['name'].isin(bsub['name'])]
    else:
        esub = exp_m

    if madf < 1:
        sys.exit("Choose MAD threshold >= 1")
    esub = dat.filter_uninformative(esub, madf)
    mtb=esub['name'].str.contains(goi).sum()

    if mtb <= 0:
        sys.exit("Gene did not pass the MAD threshold. Try setting a "
        "lower value for -m/--mt, >= 1")
    esub = esub.dropna(how='any')
    esub.set_index('name', inplace=True)
    esub.index.name = None

    if docor:
        gcor = corf
        gcor.sort_values(by=['cor'], inplace=True, ascending=False)
        gcor.index = gcor['name']
        gcor.index.name = None
        gcor.to_csv("AC.csv", index=False, header=True)
        if cort < 0.6:
            sys.exit("Choose correlation threshold >= 0.6, non-negative")
        gcor = gcor[abs(gcor['cor']) > cort]

        if len(gcor.index) < 19:
            sys.exit("You don't have enough genes that show correlation "
            "above your set threshold. Try setting a lower correlation threshold")

        if len(gcor.index) > 50:
            gcor = gcor.iloc[0:50,:]
        gcor.to_csv("TopCC.csv", index=False, header=True)

        print("Finished extracting top correlated genes")
        return gcor

    else:
        gcor = esub.corrwith(esub.loc[goi], axis=1, drop=False, method='pearson')
        gde = cont_method(esub, goi)
        gcor = gcor.to_frame()
        gcor.columns = ['cor']
        gcor['name'] = gcor.index
        gcor = gcor[['name', 'cor']]
        cor_de = pd.concat([gcor, gde], axis=1)
        gcor.sort_values(by=['cor'], inplace=True, ascending=False)
        cor_de.sort_values(by=['cor'], inplace=True, ascending=False)
        cor_de.to_csv("AC.csv", index=False, header=True)

        if cort < 0.6:
            sys.exit("Choose correlation threshold >= 0.6")
        gcor = gcor[abs(gcor['cor']) > cort]
        if len(gcor.index) < 19:
            sys.exit("You don't have enough genes that show correlation "
                     "above your set threshold. Try a lower threshold")
        if len(gcor.index) > 50:
            gcor = gcor.iloc[0:50,:]
        gcor.to_csv("TopCC.csv", index=False, header=True)
        return gcor

def make_prior1(npc, gcor):
    """
    Function to make prior matrix for ACSNI-run 1

    Parameters
    ----------
    npc: Number of user specified prior matrix columns
    gcor: Correlation Matrix

    Returns
    -------
    prior1: Prior Matrix

    """

    cc = ['1', '2', '3', '4', '5']
    for i in range(6, npc + 1):
        cc.append(str(i))
    df = pd.DataFrame(np.random.randint(1, 2, size=(len(gcor.index),
                                                    int(len(cc)))), columns=cc)
    sd = 1
    ed = 3
    for i in cc:
        df.loc[sd:ed, i] = 0
        sd = ed + 1
        ed = ed + 3
    df1 = df.copy()
    df1.loc[0, :] = 0
    df.columns = ['V' + str(col) for col in df.columns]
    df1.columns = ['repV' + str(col) for col in df1.columns]
    prior1 = pd.concat([df, df1], axis=1)
    prior1.set_index(gcor.iloc[:, 0], inplace=True)
    prior1.index.name = None
    prior1.insert(0, 'name', prior1.index)
    return prior1

def preprocess_run_one(pf, biotypefilter=False, biotypef=False, exclude=None):
    """
    Function to pre-process prediction output of ACSNI-run1

    Parameters
    ------------
    pf: Prediction output file from ACSNI-run1
    biotypef: Gene Biotype file
    biotypefilter: If biotype file and filter parameters provided.
    exclude: Biotype to exclude from Biotype file

    Returns
    -------
    predicf: Pre-processed prediction output

    """

    predicf = pd.DataFrame(data=pf)
    predicf.set_index('name', inplace=True)
    predicf.index.name = None
    predicf = predicf[predicf.sum(axis=1) >= 1]
    predicf.iloc[predicf > 0] = 1

    if biotypefilter:
        bsub2 = biotypef[~biotypef['biotype'].isin(exclude)]
        predicf = predicf[predicf.index.isin(bsub2['name'])]
    predicf.sort_index(inplace=True)
    return predicf

def process_run_one(x, y):
    """
    Function to find high confidence predicted genes from the pre-processed
    prediction output of ACSNI-run1.

    Parameters
    -----------
    x: ACSNI prediction results for option -f 0
    y: Length of initial prior

    Returns
    -------
    pdm: Processed results

    """

    pdm = list()

    for i in range(1, y + 1):
        m = "V" + str(i)
        r = "repV" + str(i)
        n = m + "vs" + r
        pdf = pd.DataFrame(x[m] > x[r], columns=[n])
        pdf = pdf[pdf[n].astype(str).str.contains("True")]
        pdm.append(pdf)

    mse = pdm[0]

    for i in pdm[1:]:
        mse = pd.concat([mse, i], axis=1)
    pdm = mse.fillna(0)
    pdm["sum"] = pdm.eq(True).sum(axis=1)
    pdm["sum"] = pdm["sum"].astype(int)
    pdm["prop"] = (pdm["sum"] / y) * 100
    pdm = pdm[pdm["prop"] >= 60]
    pdm = pdm.sort_values(by=["prop"], ascending=False)
    if len(pdm.index) > 80:
        pdm = pdm.iloc[0:80,:]
    return pdm

def make_prior2(hc, gt):
    """
    Function to make prior matrix 2 for ACSNI-run 2

    Parameters
    ----------
    hc: High-confidence predicted genes from ACSNI-run1
    gt: Gene of interest

    Returns
    -------
    prior2: Prior Matrix 2

    """
    cc2 = ['V1', 'V2']
    prior2 = pd.DataFrame(np.random.randint(1, 2, size=(len(hc.index),
    int(len(cc2)))), columns=cc2)
    prior2.set_index(hc.index, inplace=True)
    prior2.index.name = None
    goiindex = [gt]
    goi = pd.DataFrame([[1, 0]], columns=cc2, index=goiindex)
    prior2 = goi.append(prior2)
    prior2.insert(0, 'name', prior2.index)
    return prior2

def preprocess_run_two(pf2):
    """
    Function to pre-process prediction output of ACSNI-run2

    Parameters
    ------------
    pf2: Prediction output file from ACSNI-run2

    Returns
    -------
    predicf2: Pre-processed prediction output of ACSNI-run2

    """
    predicf2 = pd.DataFrame(data=pf2)
    predicf2.set_index('name', inplace=True)
    predicf2.index.name = None
    predicf2 = predicf2[predicf2.sum(axis=1) >= 1]
    predicf2.iloc[predicf2 > 0] = 1
    return predicf2

def process_run_two(hc2):
    """
    Function to find high confidence predicted genes from the pre-processed
    prediction output of ACSNI-run2.

    Parameters
    -----------
    hc2: High-confidence predicted genes from ACSNI-run2

    Returns
    -------
    pdm2: Processed results

    """
    pdm2 = pd.DataFrame(hc2['V1'] > hc2['V2'], columns=["V1vsV2"])
    pdm3 = pdm2
    pdm2 = pdm2[pdm2['V1vsV2'].astype(str).str.contains('True')]
    pdm2.columns.values[0] = "predict"
    pdm2.to_csv("predictions.csv", index=True, header=True)
    pdm3.to_csv("FD.csv", index=True, header=True)
    return pdm2

def merge_multi(path):
    """
    Merge output from multiple De-novo genesets

    Parameters
    ----------
    path: Path to the working directory

    Returns
    -------
    out: List of dataframes D, N, co files
    """
    files = [f for f in os.listdir(path) if f.endswith(".csv")]
    nn, dd, co, ac, fd = list(), list(), list(), 0, 0

    for i in files:
        if i.startswith("N"):
            nn.append(pd.read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        elif i.startswith("D"):
            dd.append(pd.read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        elif i.startswith("AC"):
            ac = pd.read_csv(os.path.join(path, i))
            os.remove(os.path.join(path, i))
        elif i.startswith("FD"):
            fd = pd.read_csv(os.path.join(path, i))
            os.remove(os.path.join(path, i))
        elif i.startswith("code"):
            co.append(pd.read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        else:
           pass

    nn = [i.set_index('name') for i in nn]
    nn = pd.concat(nn, axis=1)

    dd = [i.set_index('name') for i in dd]
    dd = pd.concat(dd, axis=1)

    co = pd.concat(co, axis=1)

    return nn, dd, ac, fd, co

def save_merged_n_d_ac_fd_co(n, d, ac, fd, co, path, nfile, run_info):
    """
    Save multi-run

    Parameters
    ----------
    n: network
    d: direction
    ac: correlations
    fd: full result dataset
    co: code
    path: Path to saved file
    nfile: Name of the impute file
    run_info: Run arguments

    Returns
    -------
    """

    dbs_results = dbs.ACSNIDeriveResults(ac=ac, n=n, d=d, fd=fd, co=co,
                                   run_info=run_info)


    mf = [f for f in os.listdir(path) if f.endswith(".h5")]
    for m in mf:
        os.remove(os.path.join(path, m))

    dbs.save_acsni(dbs_results, ("dbs" + nfile[:-4] + ".ptl"))
    return
